#!/usr/bin/env python3
"""
Merge experiments into a common directory.
"""

import argparse
import logging
import pathlib
import sys
import tempfile

import mlflow
from mlflow.client import MlflowClient

from .common import configure_logging

LOGGER = logging.getLogger(__name__)


def page_iter(func, **kwargs):
    """
    Wrapper function to transform various MLflow paged iterators into a
    continuous generator.

    Args:
        func:
            The function to call. It must accept a page_token keyword parameter.

        **kwargs:
            The keyword arguments to pass to the function.

    Returns:
        A generator over the items return by the function.
    """
    token = kwargs.get("token")
    while True:
        result = func(**kwargs)
        yield from result
        token = result.token
        if token is None:
            return


class Merger:
    """
    Merge experiments and runs from one backend into another. This uses the
    mlflow.client.MlflowClient class which is a wrapper around the MLflow REST
    API.

    * https://mlflow.org/docs/latest/api_reference/python_api/mlflow.client.html
    * https://mlflow.org/docs/latest/api_reference/rest-api.html
    """

    def __init__(self, src_kwargs, dst_kwargs):
        """
        Args:
            src_kwargs:
                Keyword arguments for MlflowClient for the source client.

            dst_kwargs:
                Keyword arguments for MlflowClient for the destination client.
        """
        self.src_client = MlflowClient(**src_kwargs)
        self.dst_client = MlflowClient(**dst_kwargs)
        self.exp_ids = {}

    def copy_experiments(self):
        """
        Copy experiments.
        """
        for exp in page_iter(self.src_client.search_experiments):
            existing_exp = self.dst_client.get_experiment_by_name(exp.name)
            # Upload the experiment if it is missing.
            if existing_exp is None:
                LOGGER.info("Copying experiment: %s", exp.name)
                dst_id = self.dst_client.create_experiment(
                    name=exp.name,
                    artifact_location=exp.artifact_location,
                    tags=exp.tags,
                )
            else:
                LOGGER.info("Found experiment: %s", exp.name)
                dst_id = existing_exp.experiment_id
            self.exp_ids[exp.experiment_id] = dst_id

    def copy_datasets(self):
        """
        Copy datasets.
        """
        raise NotImplementedError("copy_datasets")

        if not hasattr(self.dst_client, "create_dataset"):
            LOGGER.warning(
                "The create_dataset method has been removed: datasets will not be copied"
            )
            return

        exp_ids = list(self.exp_ids)
        for dataset in page_iter(
            self.src_client.search_datasets, experiment_ids=exp_ids
        ):
            # TODO
            # Check if the name is the same as the dataset_id.
            existing_dataset = self.dst_client.get_dataset(dataset.name)
            if existing_dataset is None:
                new_dataset = self.dst_client.create_dataset(
                    name=dataset.name,
                    experiment_id=[self.exp_ids[e] for e in dataset.experiment_ids],
                    tags=dataset.tags,
                )
                # TODO
                # Transfer the data itself.
            # TODO
            # Create a dataset map if the IDs can differ.

    def copy_runs(self):
        """
        Copy runs.
        """
        exp_ids = list(self.exp_ids)
        for run in page_iter(self.src_client.search_runs, experiment_ids=exp_ids):
            try:
                self.dst_client.get_run(run.info.run_id)
                LOGGER.info("Found run: %s", run.info.run_id)
                continue
            # An exception occurs if the run does not exist.
            except mlflow.exceptions.MlflowException as err:
                # Check if the error includes the run ID. If not, something else
                # went wrong so reraise the exception.
                if run.info.run_id not in str(err):
                    raise err from err
            LOGGER.info("Copying run: %s", run.info.run_id)
            new_run = self.dst_client.create_run(
                experiment_id=self.exp_ids[run.info.experiment_id],
                start_time=run.info.start_time,
                tags=run.data.tags,
                run_name=run.info.run_name,
            )
            metrics = []
            for metric_name in run.data.metrics:
                metrics.extend(
                    self.src_client.get_metric_history(run.info.run_id, metric_name)
                )
            self.dst_client.log_batch(
                run_id=new_run.info.run_id,
                metrics=metrics,
                params=(
                    mlflow.entities.Param(k, v) for (k, v) in run.data.params.items()
                ),
            )
            self.copy_run_artifacts(run, new_run)

    def copy_run_artifacts(self, src_run, dst_run):
        """
        Copy run artifacts.
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            for artifact in self.src_client.list_artifacts(src_run.info.run_id):
                LOGGER.info(
                    "Copying artifact: %s [run_id: %s]",
                    artifact.path,
                    src_run.info.run_id,
                )
                tmp_path = tmp_dir / artifact.path
                self.src_client.download_artifacts(
                    run_id=src_run.info.run_id, path=artifact.path, dst_path=tmp_path
                )
                self.dst_client.log_artifact(dst_run.info.run_id, tmp_path)

    # https://mlflow.org/docs/latest/ml/model-registry
    def copy_models(self):
        """
        Copy logged models.
        """
        exp_ids = list(self.exp_ids)
        dst_exp_ids = list(self.exp_ids.values())
        for model in page_iter(
            self.src_client.search_logged_models, experiment_ids=exp_ids
        ):
            existing_model = self.dst_client.search_logged_models(
                experiment_ids=dst_exp_ids, filter_string=f"name = {model.name}"
            )
            if existing_model:
                LOGGER.info("Found model: %s", model.name)
                continue
            new_model = self.dst_client.create_logged_model(
                experiment_id=self.exp_ids[model.experiment_id],
                name=model.name,
                source_run_id=model.source_run_id,
                tags=model.tags,
                params=model.params,
                model_type=model.type,
            )
            self.copy_model_artifacts(model, new_model)
            # create_model_version
            # set_registered_model_alias

    def copy_model_artifacts(self, src_model, dst_model):
        """
        Copy model artifacts.
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            for artifact in self.src_client.list_logged_model_artifacts(
                src_model.model_id
            ):
                LOGGER.info(
                    "Copying artifact: %s [model name: %s]",
                    artifact.path,
                    src_model.name,
                )
                tmp_path = tmp_dir / artifact.path
                self.src_client.download_artifacts(
                    run_id=src_model.source_run_id,
                    path=artifact.path,
                    dst_path=tmp_path,
                )
                self.dst_client.log_model_artifact(dst_model.model_id, tmp_path)

    def copy_all(self):
        """
        Copy everything that is currently supported.
        """
        self.copy_experiments()
        self.copy_runs()
        self.copy_models()


def script_merge_backend(args=None):
    """
    Copy experiments and runs from one backend to another. This is a work in
    progress and should be used with caution.
    """
    configure_logging()

    parser = argparse.ArgumentParser(description=script_merge_backend.__doc__)
    parser.add_argument(
        "--tua",
        help="Tracking URI for the source backend.",
    )
    parser.add_argument(
        "--rua",
        help="Registry URI for the source backend.",
    )
    parser.add_argument(
        "--wua",
        help="Workspace store URI for the source backend.",
    )
    parser.add_argument(
        "--tub",
        help="Tracking URI for the destination backend.",
    )
    parser.add_argument(
        "--rub",
        help="Registry URI for the destination backend.",
    )
    parser.add_argument(
        "--wub",
        help="Workspace store URI for the destination backend.",
    )
    pargs = parser.parse_args(args=args)

    merger = Merger(
        src_kwargs={
            "tracking_uri": pargs.tua,
            "registry_uri": pargs.rua,
            "workspace_store_uri": pargs.wua,
        },
        dst_kwargs={
            "tracking_uri": pargs.tub,
            "registry_uri": pargs.rub,
            "workspace_store_uri": pargs.wub,
        },
    )
    try:
        merger.copy_all()
    except OSError as err:
        sys.exit(err)


if __name__ == "__main__":
    script_merge_backend()



#!/usr/bin/env python3
"""
URI manipulation.
"""

import pathlib
import urllib.parse

FILENAME = "meta.yaml"


class URITransformer:
    """
    Transform URIs, including file paths.
    """

    def __init__(self, uri):
        self.parts = urllib.parse.urlsplit(uri)

    def __str__(self):
        return urllib.parse.urlunsplit(self.parts)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def copy(self, path=None):
        """
        Get a copy of this instance.
        """
        cpy = self.__class__(str(self))
        if path is not None:
            cpy.path = path
        return cpy

    @property
    def scheme(self):
        """
        The URI scheme.
        """
        return self.parts.scheme

    @property
    def is_local_file(self):
        """
        True if the URI points to a local path, else False. The path may not exist.
        """
        scheme = self.scheme
        return (not scheme) or scheme.lower() == "file"

    @property
    def path(self):
        """
        The path.
        """
        return pathlib.Path(self.parts.path)

    @path.setter
    def path(self, path):
        if self.is_local_file and self.path.is_absolute():
            path = path.resolve()
        # Despite the understore, _replace is supported in the official
        # documentation.
        self.parts = self.parts._replace(path=str(path))



#!/usr/bin/env python3
"""
Common functionality.
"""

import logging

DEFAULT_DIRECTORY = "mlruns"


def configure_logging(level=logging.INFO):
    """
    Configure logging.

    Args:
        level:
            The logging level.
    """
    logging.basicConfig(
        style="{",
        format="[{asctime}] {levelname} {message}",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=level,
    )



#!/usr/bin/env python3
"""Limited interace to MLflow's SQLite database."""

import pathlib
import sqlite3


class MLflowSqlite:
    """
    Limited interface to MLflow's SQLite database.
    """

    DEFAULT_PATH = "mlflow.db"

    def __init__(self, path=DEFAULT_PATH):
        self.path = pathlib.Path(path).resolve()
        self._conn = None

    @property
    def conn(self):
        """
        The connection.
        """
        return sqlite3.connect(self.path)

    def update_artifact_uris(self, artifact_map):
        """
        Update artifact URIs in the database.

        Args:
            artifact_map:
                A dict mapping run IDs to new artifact URIs.
        """
        args = [(v, k) for (k, v) in artifact_map.items()]
        with self.conn as conn:
            conn.executemany(
                "UPDATE runs SET artifact_uri = ? WHERE run_uuid = ?", args
            )



#!/usr/bin/env python3
"""Functions for managing artifacts."""

import argparse
import functools
import logging
import pathlib
import sys

import mlflow
import yaml

from .common import configure_logging
from .sqlite import MLflowSqlite
from .uri import URITransformer

LOGGER = logging.getLogger(__name__)


def load_map(path):
    """
    Load a YAML file that maps old artifact URIs to new ones.
    """
    path = pathlib.Path(path).resolve()
    with path.open("rb", encoding="utf-8") as handle:
        data = yaml.safe_load(handle)
    uri_map = {}
    for old, new in data.items():
        old = URITransformer(old)
        new = URITransformer(new)
        uri_map[old] = new
    return uri_map


class ArtifactURIMapper:
    """
    Map old local articate URIs to new ones.
    """

    def __init__(self, paths=None, map_path=None):
        """
        Args:
            paths:
                An optional iterable over mlruns-like directories in which to
                search for displaced artifacts.

            map_path:
                A path to a YAML file containing a dict that maps old artifact
                URIs to new ones.
        """
        self.paths = paths
        self.map_path = map_path

    @functools.cached_property
    def user_map(self):
        """
        A dict mapping old URIs to new ones.
        """
        if self.map_path:
            return load_map(self.map_path)
        return {}

    @functools.cached_property
    def user_mapped_paths(self):
        """
        The reverse-sorted list of mapped URIs if user_map is not None, else
        None. The first match when searching for relative subpaths will be the
        most specific due to the sort order.
        """
        return sorted((u.path for u in self.user_map), reverse=True)

    def get_new_artifact_uri(self, artifact_uri, run):
        """
        Get the new URI for the given artifact URI.
        """
        # If there is a direct mapping, use it.
        if artifact_uri in self.user_map:
            return self.user_map[artifact_uri]

        # If the URI is not a local file, skip it.
        if not artifact_uri.is_local_file:
            return None

        artifact_path = artifact_uri.path

        for old in self.user_mapped_paths:
            if old.is_local_file and artifact_path.is_relative_to(old):
                path = self.user_map[old] / artifact_path.relative_to(old)
                return artifact_uri.copy(path=path)

        parts = artifact_path.parts
        expected_parts = (run.info.experiment_id, run.info.run_id, "artifacts")
        if parts[-3:] != expected_parts:
            LOGGER.error(
                "Expected current artifact path to end with %s, not %s",
                expected_parts,
                parts[-3:],
            )
            return None

        subpath = pathlib.Path(*expected_parts)
        for p in self.paths:
            path = p / subpath
            if path.exists():
                return artifact_uri.copy(path=path)

        return None

    def map_artifact_uris(self):
        """
        Get a dict mapping old artifact URIs to new ones.
        """
        artifact_map = {}
        for run in mlflow.search_runs(
            experiment_ids=[e.experiment_id for e in mlflow.search_experiments()],
            output_format="list",
        ):
            artifact_uri = URITransformer(run.info.artifact_uri)
            new_uri = self.get_new_artifact_uri(artifact_uri, run)

            if new_uri is None:
                if artifact_uri.is_local_file and not artifact_uri.path.exists():
                    LOGGER.warning(
                        "Local artifact path could not be located: %s [run_id: %s]",
                        artifact_uri,
                        run.info.run_id,
                    )
            elif artifact_uri != new_uri:
                LOGGER.info(
                    "Artifact URI will be updated from %s to %s [run_id: %s]",
                    artifact_uri,
                    new_uri,
                    run.info.run_id,
                )
                artifact_map[run.info.run_id] = str(new_uri)

        return artifact_map


def script_fix_artifact_uris(args=None):
    """
    Attempt to fix broken artifact URIs in experiments and runs. This only works
    if the artifacts can be found in local mlruns directories or directories
    with equivalent structures, and when the MLflow backend is a local SQLite3
    database.
    """
    configure_logging()

    parser = argparse.ArgumentParser(description=script_fix_artifact_uris.__doc__)
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        type=pathlib.Path,
        help="Paths to directories containingartifacts, such as an mlruns directory.",
    )
    parser.add_argument(
        "-m", "--map", help="A path to a YAML file that maps old paths to new paths."
    )
    parser.add_argument(
        "--db",
        "--database",
        type=pathlib.Path,
        default=MLflowSqlite.DEFAULT_PATH,
        help="Path to the local MLflow SQLite database. Default: %(default)s",
    )
    try:
        pargs = parser.parse_args(args=args)
        uri_map = ArtifactURIMapper(
            paths=pargs.paths, map_path=pargs.map
        ).map_artifact_uris()
        if uri_map:
            MLflowSqlite(path=pargs.db).update_artifact_uris(uri_map)
    except OSError as err:
        sys.exit(err)
